#include "alsa.h"

#include <stdio.h>
#include <alsa/asoundlib.h>

#define A do { if (err < 0) { fprintf(stderr, "alsa-out-xrun: %s:%d: alsa error '%s'\n", __FILE__, __LINE__, snd_strerror(err)); exit(1); } } while (0)

void *alsa_open(void)
{
  alsa_data           *a;
  snd_pcm_t           *h;
  snd_pcm_hw_params_t *hw_params;
  snd_pcm_sw_params_t *sw_params;
  snd_pcm_uframes_t   period_frames;
  snd_pcm_uframes_t   buffer_frames;
  unsigned int        rate;
  snd_pcm_format_t    format;
  int err;

  a = malloc(sizeof(alsa_data)); if (a == NULL) { fprintf(stderr, "out of memory\n"); exit(1); }
  a->samplerate = 48000;
  a->channels = 2;
  a->device = "default";

  format = SND_PCM_FORMAT_S16_LE;
  err = snd_pcm_open(&h, a->device, SND_PCM_STREAM_PLAYBACK, 0); A;
  snd_pcm_hw_params_alloca(&hw_params);
  snd_pcm_sw_params_alloca(&sw_params);
  err = snd_pcm_hw_params_any(h, hw_params); A;
  err = snd_pcm_hw_params_set_access(h, hw_params, SND_PCM_ACCESS_RW_INTERLEAVED); A;
  err = snd_pcm_hw_params_set_format(h, hw_params, format); A;
  err = snd_pcm_hw_params_set_channels(h, hw_params, a->channels); A;
  rate = a->samplerate;
  err = snd_pcm_hw_params_set_rate_near(h, hw_params, &rate, 0); A;
  if (rate != a->samplerate) {fprintf(stderr, "bad rate\n"); exit(1);}
  /* dumb values */
  period_frames = 1024; //a->nsamples;
  buffer_frames = 3 * 1024;   //period_frames * a->nperiods;
  err = snd_pcm_hw_params_set_period_size_near(h,hw_params,&period_frames,0);A;
  err = snd_pcm_hw_params_set_buffer_size_near(h, hw_params, &buffer_frames);A;
  err = snd_pcm_hw_params(h, hw_params); A;
//  err = snd_pcm_start(h); A;
  a->h = h;
  return a;
}

int alsa_latency(void *_a)
{
  alsa_data *a = _a;
  snd_pcm_sframes_t ret;
  int err = snd_pcm_delay(a->h, &ret);
  if (err) ret = 0;
  /* useless test? can it be <0? */
  if (ret < 0) ret = 0;
  return ret;
}

static void xrun(alsa_data *a)
{
  char data[1024*2*2];
  fprintf(stderr, "xrun\n");
  memset(data, 0, 1024*2*2);
  snd_pcm_prepare(a->h);
  snd_pcm_writei(a->h, data, 1024);
}

static void suspend(alsa_data *a)
{
  int res;
  int loop = 0;
  fprintf(stderr, "suspend\n");
  while ((res = snd_pcm_resume(a->h)) == -EAGAIN && loop < 100) {
    struct timeval t;
    t.tv_sec = 0;
    t.tv_usec = 10 * 1000;
    select(0, 0, 0, 0, &t);
    loop++;
  }
  if (res < 0) xrun(a); //snd_pcm_prepare(a->h);
}

void alsa_play(void *_a, char *buf, int nsamples)
{
  alsa_data *a = _a;
#define X do { fprintf(stderr, "alsa-out-xrun: %s:%d: alsa error\n", __FILE__, __LINE__); xrun(a); } while (0)
#define Y do { fprintf(stderr, "alsa-out-xrun: %s:%d: alsa error\n", __FILE__, __LINE__); suspend(a); } while (0)
  size_t count = nsamples;
  ssize_t r;
  unsigned char *bb;

  bb = (unsigned char *)buf;

  while (count) {
    r = snd_pcm_writei(a->h, bb, count);
    if (r == -EAGAIN || (r >= 0 && (size_t)r < count)) {
      /* nothing */
    }
    else if (r == -EPIPE) X;
    else if (r == -ESTRPIPE) Y;
    else if (r < 0) X;
    if (r > 0) {
      count -= r;
      bb += r * a->channels * 2;
    }
  }
#undef X
}

void alsa_drop(void *_a)
{
  alsa_data *a = _a;
  if (snd_pcm_drop(a->h)) { fprintf(stderr, "alsa_drop fails\n"); exit(1); }
}

void alsa_drain(void *_a)
{
  alsa_data *a = _a;
  if (snd_pcm_drain(a->h)) { fprintf(stderr, "alsa_drain fails\n"); exit(1); }
}
