#include "conf.h"
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>

static char conf_file[FILENAME_MAX+1];

typedef struct {
  char name[128];
  float val;
} conf_item;

static conf_item *conf;
static int confsize;
static int confmaxsize;

static void save_conf(void)
{
  int i;
  FILE *f = fopen(conf_file, "w");
  if (f == NULL) return;
  for (i = 0; i < confsize; i++)
    fprintf(f, "%s %g\n", conf[i].name, conf[i].val);
  fclose(f);
}

static void get_conf_line(FILE *f, char *s)
{
  int i = 0;
  int c;
empty_line:
  while (1) {
    c = fgetc(f); if (c == EOF || c == '\n') break;
    if (isspace(c)) c = ' ';
    if (c == '#') { while (c != EOF && c != '\n') c = fgetc(f); break; }
    s[i] = c;
    i++; if (i == 1024) abort();
  }
  if (i == 0 && c == '\n') goto empty_line;
  s[i] = 0;
}

/* duplicate values not detected */
static void add_conf(char *s, float val)
{
  if (strlen(s) > 127) abort();
  if (confsize == confmaxsize) {
    confmaxsize += 16;
    conf = realloc(conf, confmaxsize * sizeof(conf_item));
    if (conf == NULL) abort();
  }
  strcpy(conf[confsize].name, s);
  conf[confsize].val = val;
  confsize++;
}

static void load_conf(void)
{
  FILE *f;
  free(conf);
  conf = NULL;
  confsize = confmaxsize = 0;
  f = fopen(conf_file, "r"); if (f == NULL) return;
  while (1) {
    char s[1024], *s2;
    get_conf_line(f, s);
    if (s[0] == 0) break;
    s2 = strchr(s, ' '); if (s2 == NULL) abort();
    *s2 = 0;
    s2++;
    add_conf(s, atof(s2));
  }
  fclose(f);
}

void init_conf(char *file)
{
  strcpy(conf_file, file);
  load_conf();
}

conf_val get_conf(char *s)
{
  conf_val ret;
  int i;
  for (i = 0; i < confsize; i++) if (!strcmp(s, conf[i].name)) {
    ret.type = FLOAT;
    ret.val = conf[i].val;
    return ret;
  }
  ret.type = NONE;
  return ret;
}

void store_conf_float(char *name, float v)
{
  int i;
  for (i = 0; i < confsize; i++) if (!strcmp(conf[i].name, name)) break;
  if (i == confsize) add_conf(name, v);
  else conf[i].val = v;
  save_conf();
}
