#include "downloader.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

typedef struct {
  int video_pipe;
  int audio_pipe;
} downloader_internal_t;

/* I'm not a big fan of this __USE_GNU, it's needed to have F_SETPIPE_SZ */
/* we can't include both fnctl.h and linux/fcntl.h, go figure... */
#define __USE_GNU
#include <fcntl.h>

/* I keep this here, just in case... */
#ifndef F_SETPIPE_SZ
#define F_SETPIPE_SZ 1031
#endif

downloader_t *new_downloader(int video_pipe, int audio_pipe)
{
  downloader_internal_t *ret = calloc(1, sizeof(*ret)); if (ret == NULL) abort();

  ret->video_pipe = video_pipe;

  int size = 0x800000;
  size = fcntl(ret->video_pipe, F_SETPIPE_SZ, size);
  printf("video size %d\n", size);

  ret->audio_pipe = audio_pipe;

  size = 48000;
  size = fcntl(ret->audio_pipe, F_SETPIPE_SZ, size);
  printf("audio size %d\n", size);

  return ret;
}

static char *download(int pipe, int size)
{
  char *ret = malloc(size); if (ret == NULL) abort();
  int pos = 0;

  while (size) {
    int l = read(pipe, ret + pos, size);
    if (l <= 0) {
      free(ret);
      return NULL;
    }
    size -= l;
    pos += l;
  }

  return ret;
}

char *download_video(downloader_t *_d, int size)
{
  downloader_internal_t *d = _d;
  return download(d->video_pipe, size);
}

char *download_audio(downloader_t *_d, int size)
{
  downloader_internal_t *d = _d;
  return download(d->audio_pipe, size);
}

void delete_downloader(downloader_t *dl)
{
  free(dl);
}
