/* stolen from Adriaensen FIL plugin */
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "eq.h"
#include "conf.h"

static float exp2ap (float x)
{
    int i;

    i = (int)(floorf (x));
    x -= i;
    return ldexpf (1 + x * (0.6930f + x * (0.2416f + x * (0.0517f + x * 0.0137f))), i);
}

enum { Paramsect_SECT, Paramsect_FREQ, Paramsect_BAND, Paramsect_GAIN };

typedef struct {
    float  _f, _b, _g;
    float  _s1, _s2, _a;
    float  _z1, _z2;
} Paramsect;

static void Paramsect_init(Paramsect *this)
{
        this->_f = 0.25f;
        this->_b = this->_g = 1.0f;
        this->_a = this->_s1 = this->_s2 = this->_z1 = this->_z2 = 0.0f;
}

static void Paramsect_proc(Paramsect *this, int k, float *sig, float f, float b, float g)
{
        float s1, s2, d1, d2, a, da, x, y;
        int  u2 = 0;

        s1 = this->_s1;
        s2 = this->_s2;
        a = this->_a;
        d1 = 0;
        d2 = 0;
        da = 0;

        if (f != this->_f)
        {
            if      (f < 0.5f * this->_f) f = 0.5f * this->_f;
            else if (f > 2.0f * this->_f) f = 2.0f * this->_f;
            this->_f = f;
            this->_s1 = -cosf (6.283185f * f);
            d1 = (this->_s1 - s1) / k;
            u2 = 1;
        }

        if (g != this->_g)
        {
            if      (g < 0.5f * this->_g) g = 0.5f * this->_g;
            else if (g > 2.0f * this->_g) g = 2.0f * this->_g;
            this->_g = g;
            this->_a = 0.5f * (g - 1.0f);
            da = (this->_a - a) / k;
            u2 = 1;
        }

        if (b != this->_b)
        {
            if      (b < 0.5f * this->_b) b = 0.5f * this->_b;
            else if (b > 2.0f * this->_b) b = 2.0f * this->_b;
            this->_b = b;
            u2 = 1;
        }

        if (u2)
        {
            b *= 7 * f / sqrtf (g);
            this->_s2 = (1 - b) / (1 + b);
            d2 = (this->_s2 - s2) / k;
        }

        while (k--)
        {
            s1 += d1;
            s2 += d2;
            a += da;
            x = *sig;
            y = x - s2 * this->_z2;
            *sig++ -= a * (this->_z2 + s2 * y - x);
            y -= s1 * this->_z1;
            this->_z2 = this->_z1 + s1 * y;
            this->_z1 = y + 1e-10f;
        }
}

enum { Ladspa_Paramfilt_AIP, Ladspa_Paramfilt_AOP, Ladspa_Paramfilt_FILT, Ladspa_Paramfilt_GAIN, Ladspa_Paramfilt_SECT, Ladspa_Paramfilt_NSECT = 4, Ladspa_Paramfilt_NPORT = 20 };

typedef struct {
    float     *_port [Ladspa_Paramfilt_NPORT];
    float      _gain;
    int        _fade;
    Paramsect  _sect [Ladspa_Paramfilt_NSECT];
  float _fsam;
} Ladspa_Paramfilt;

#if 0
static void Ladspa_Paramfilt_setgain(Ladspa_Paramfilt *this, float gain)
{
  this->_gain = gain;
}
#endif

static void Ladspa_Paramfilt_setport(Ladspa_Paramfilt *this,
    unsigned long port, void *data)
{
    this->_port [port] = data;
}

static void Ladspa_Paramfilt_active(Ladspa_Paramfilt *this, int act)
{
    int j;

    if (! act) return;
    this->_fade = 0;
    this->_gain = 1;
    for (j = 0; j < Ladspa_Paramfilt_NSECT; j++) Paramsect_init(&this->_sect[j]);
}

static void Ladspa_Paramfilt_runproc( Ladspa_Paramfilt *this,
    unsigned long len, int add)
{
    int   i, j, k;
    float *aip = this->_port [Ladspa_Paramfilt_AIP];
    float *aop = this->_port [Ladspa_Paramfilt_AOP];
    float *p, sig [48];
    float t, g, d;
    float fgain;
    float sfreq [Ladspa_Paramfilt_NSECT];
    float sband [Ladspa_Paramfilt_NSECT];
    float sgain [Ladspa_Paramfilt_NSECT];

    fgain = exp2ap (0.1661 * this->_port [Ladspa_Paramfilt_GAIN][0]);
    for (j = 0; j < Ladspa_Paramfilt_NSECT; j++)
    {
        t = this->_port [Ladspa_Paramfilt_SECT + 4 * j + Paramsect_FREQ][0] / this->_fsam;
        if (t < 0.0002) t = 0.0002;
        if (t > 0.4998) t = 0.4998;
        sfreq [j] = t;
        sband [j] = this->_port [Ladspa_Paramfilt_SECT + 4 * j + Paramsect_BAND][0];
        if (this->_port [Ladspa_Paramfilt_SECT + 4 * j + Paramsect_SECT][0] > 0) sgain [j] = exp2ap (0.1661 * this->_port [Ladspa_Paramfilt_SECT + 4 * j + Paramsect_GAIN][0]);
        else sgain [j] = 1.0;
    }

    while (len)
    {
        k = (len > 48) ? 32 : len;

        t = fgain;
        g = this->_gain;
        if      (t > 1.25 * g) t = 1.25 * g;
        else if (t < 0.80 * g) t = 0.80 * g;
        this->_gain = t;
        d = (t - g) / k;
        for (i = 0; i < k; i++)
        {
            g += d;
            sig [i] = g * aip [i];
        }

        for (j = 0; j < Ladspa_Paramfilt_NSECT; j++) Paramsect_proc(&this->_sect [j], k, sig, sfreq [j], sband [j], sgain [j]);

        j = this->_fade;
        g = j / 16.0;
        p = 0;
        if (this->_port [Ladspa_Paramfilt_FILT][0] > 0)
        {
            if (j == 16) p = sig;
            else ++j;
        }
        else
        {
            if (j == 0) p = aip;
            else --j;
        }
        this->_fade = j;
        if (p) memcpy (aop, p, k * sizeof (float));
        else
        {
            d = (j / 16.0 - g) / k;
            for (i = 0; i < k; i++)
            {
                g += d;
                aop [i] = g * sig [i] + (1 - g) * aip [i];
            }
        }
        aip += k;
        aop += k;
        len -= k;
    }
}

#define LOAD_CONF(v, def, name) \
  do { \
    char s[128]; \
    conf_val val; \
    sprintf(s, name, i); \
    val = get_conf(s); \
    if (val.type == NONE) v = def; else v = val.val; \
  } while (0)
#define LOAD_CONF1(v, def, name) \
  do { \
    conf_val val; \
    val = get_conf(name); \
    if (val.type == NONE) v = def; else v = val.val; \
  } while (0)

void init_eq(eq *e, float f)
{
  int i;
  Ladspa_Paramfilt *p;

  memset(e, 0, sizeof(*e));

  e->samplerate = f;
  e->frequency_low[0] = 20;
  e->frequency_low[1] = 40;
  e->frequency_low[2] = 100;
  e->frequency_low[3] = 200;

  e->frequency_high[0] = 2000;
  e->frequency_high[1] = 4000;
  e->frequency_high[2] = 10000;
  e->frequency_high[3] = 20000;

  e->filter = 1;

  LOAD_CONF1(e->filter_gain, 0, "eq.gain");

  for (i = 0; i < 4; i++) {
    /* f(0) = f0 = flow
     * f(1) = f1 = fhigh
     * f(x) = a * e^(bx)     |    x = log(f(x)/a)/b
     * f0 = a
     * f1 = a * e^b
     * => f1 = f0 * e^b
     *    f1/f0 = e^b
     *    log(f1/f0) = b
     */
    float a = e->frequency_low[i];
    float b = log(e->frequency_high[i]/e->frequency_low[i]);
    LOAD_CONF(e->section[i], 0, "eq.%d.on");
    LOAD_CONF(e->frequency[i], a * exp(b * 0.5), "eq.%d.frequency");
    LOAD_CONF(e->bandwidth[i], 1, "eq.%d.bandwidth");
    LOAD_CONF(e->gain[i], 0, "eq.%d.gain");
  }

  e->paramfilt_left = malloc(sizeof(Ladspa_Paramfilt));
  if (e->paramfilt_left == NULL) abort();
  e->paramfilt_right = malloc(sizeof(Ladspa_Paramfilt));
  if (e->paramfilt_right == NULL) abort();

  memset(e->paramfilt_left, 0, sizeof(Ladspa_Paramfilt));
  memset(e->paramfilt_right, 0, sizeof(Ladspa_Paramfilt));

  p = e->paramfilt_left;
  p->_fsam = f;
  Ladspa_Paramfilt_setport(p, 0, NULL);
  Ladspa_Paramfilt_setport(p, 1, NULL);
  Ladspa_Paramfilt_setport(p, 2, &e->filter);
  Ladspa_Paramfilt_setport(p, 3, &e->filter_gain);
  for (i = 0; i < 4; i++) {
    Ladspa_Paramfilt_setport(p, i * 4 + 4, &e->section[i]);
    Ladspa_Paramfilt_setport(p, i * 4 + 5, &e->frequency[i]);
    Ladspa_Paramfilt_setport(p, i * 4 + 6, &e->bandwidth[i]);
    Ladspa_Paramfilt_setport(p, i * 4 + 7, &e->gain[i]);
  }
  Ladspa_Paramfilt_active(p, 1);

  p = e->paramfilt_right;
  p->_fsam = f;
  Ladspa_Paramfilt_setport(p, 0, NULL);
  Ladspa_Paramfilt_setport(p, 1, NULL);
  Ladspa_Paramfilt_setport(p, 2, &e->filter);
  Ladspa_Paramfilt_setport(p, 3, &e->filter_gain);
  for (i = 0; i < 4; i++) {
    Ladspa_Paramfilt_setport(p, i * 4 + 4, &e->section[i]);
    Ladspa_Paramfilt_setport(p, i * 4 + 5, &e->frequency[i]);
    Ladspa_Paramfilt_setport(p, i * 4 + 6, &e->bandwidth[i]);
    Ladspa_Paramfilt_setport(p, i * 4 + 7, &e->gain[i]);
  }
  Ladspa_Paramfilt_active(p, 1);
}

void close_eq(eq *e)
{
  free(e->paramfilt_left);
  free(e->paramfilt_right);
}

static int16_t to_short(float x)
{
  int ret;
  if (x > 1) x = 1;
  if (x < -1) x = -1;
  ret = x * 32768;
  if (ret > 32767) ret = 32767;
  return ret;
}

void run_eq(eq *e, int16_t *buf, int len, int channels)
{
  float eqin[len];
  float eqout[len];
  int i;

  if (channels != 2) abort();

  Ladspa_Paramfilt_setport(e->paramfilt_left, 0, eqin);
  Ladspa_Paramfilt_setport(e->paramfilt_left, 1, eqout);
  for (i = 0 ; i < len; i++) eqin[i] = buf[i*2] / 32768.;
  Ladspa_Paramfilt_runproc(e->paramfilt_left, len, 0);
  for (i = 0 ; i < len; i++) buf[i*2] = to_short(eqout[i]);

  Ladspa_Paramfilt_setport(e->paramfilt_right, 0, eqin);
  Ladspa_Paramfilt_setport(e->paramfilt_right, 1, eqout);
  for (i = 0 ; i < len; i++) eqin[i] = buf[i*2+1] / 32768.;
  Ladspa_Paramfilt_runproc(e->paramfilt_right, len, 0);
  for (i = 0 ; i < len; i++) buf[i*2+1] = to_short(eqout[i]);
}
