#include "ffmpeg.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <fcntl.h>

static int sub_process(int *piperet, char *prog, char **arg, char *freeme, char *pipearg)
{
  int pp[2];
  int s;

  if (pipe(pp)) abort();
  *piperet = pp[0];
  sprintf(pipearg, "pipe:%d", pp[1]);

  s = fork();

  if (s == -1) { fprintf(stderr, "fork failed\n"); exit(1); }

  if (s == 0) {
    int i;
    char **s = arg;
    printf("run: %s", prog);
    while (*s) { printf(" %s", *s); s++; }
    printf("\n");

    for (i = 0; i<100; i++) if (i != pp[1]) close(i);
    int devnull = open("/dev/null", O_RDWR);
    if (devnull == -1) { perror("/dev/null"); exit(1); }
    dup2(devnull, 0);
    dup2(devnull, 1);
    dup2(devnull, 2);
    execvp(prog, arg);
    perror("execvp");
    exit(1);
  }

  close(pp[1]);
  free(freeme);
  return s;
}

int ffmpeg_video(int *pipe, int width, int height,
                 char *video, char *subtitle, int subtitle_stream,
                 int start_second)
{
//  char *arg[23];
char *arg[128];
  char start[32];
  int p = 0;
  int h, m, s;
  char *sub = NULL;
  char vsize[32];
  char pp[32];

  s = start_second % 60;
  m = (start_second / 60) % 60;
  h = start_second / (60 * 60);
  sprintf(start, "%d:%d:%d", h, m, s);

  arg[p] = "ffmpeg";       p++;
  if (start_second) {
    arg[p] = "-ss";          p++;
    arg[p] = start;          p++;
  }
  arg[p] = "-copyts";      p++;
  arg[p] = "-i";           p++;
  arg[p] = video;          p++;
  if (subtitle != NULL) {
    int len = strlen(subtitle) + 64;
    sub = malloc(len); if (sub == NULL) abort();
    if (subtitle_stream != -1)
      sprintf(sub, "subtitles=%s:stream_index=%d", subtitle, subtitle_stream);
    else
      sprintf(sub, "subtitles=%s", subtitle);
    if (start_second) {
      arg[p] = "-ss";        p++;
      arg[p] = start;        p++;
    }
    arg[p] = "-vf";        p++;
    arg[p] = sub;          p++;
  }
  arg[p] = "-vcodec";      p++;
  arg[p] = "rawvideo";     p++;
  arg[p] = "-f";           p++;
  arg[p] = "rawvideo";     p++;
  arg[p] = "-pix_fmt";     p++;
  arg[p] = "rgba";         p++;
  arg[p] = "-video_size";  p++;
  sprintf(vsize, "%dx%d", width, height);
  arg[p] = vsize;          p++;
  //arg[p] = "-vf";           p++;
  //arg[p] = "yadif";           p++;
  arg[p] = pp;             p++;
  arg[p] = "-y";           p++;
  //arg[p] = "-loglevel";    p++;
  //arg[p] = "quiet";        p++;
  arg[p] = NULL;

  return sub_process(pipe, "ffmpeg", arg, sub, pp);
}

int ffmpeg_audio(int *pipe, char *audio, int audio_track, int start_second)
{
  char *arg[32];
  char start[32];
  char pp[32];
  char map_audio[32];
  int p = 0;
  int h, m, s;

  s = start_second % 60;
  m = (start_second / 60) % 60;
  h = start_second / (60 * 60);
  sprintf(start, "%d:%d:%d", h, m, s);

  arg[p] = "ffmpeg";       p++;
  if (start_second) {
    arg[p] = "-ss";          p++;
    arg[p] = start;          p++;
  }
  arg[p] = "-i";           p++;
  arg[p] = audio;          p++;
  arg[p] = "-f";           p++;
  arg[p] = "s16le";        p++;
  arg[p] = "-acodec";      p++;
  arg[p] = "pcm_s16le";    p++;
  arg[p] = "-ac";          p++;
  arg[p] = "2";            p++;
  arg[p] = "-ar";          p++;
  arg[p] = "48000";        p++;
  if (audio_track != -1) {
    sprintf(map_audio, "0:a:%d", audio_track);
    arg[p] = "-map";       p++;
    arg[p] = map_audio;    p++;
  }
  arg[p] = pp;             p++;
  arg[p] = "-y";           p++;
  //arg[p] = "-loglevel";    p++;
  //arg[p] = "quiet";        p++;
  arg[p] = NULL;

  return sub_process(pipe, "ffmpeg", arg, NULL, pp);
}
