#include "help.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>

#include "text_over.h"
#include "util.h"
#include "x.h"

#define gprintf(...) cgprintf(255, 255, 255, __VA_ARGS__)

static character *screen;
static int maxlines;
static int column;
static int line;
static int start_line;

static int space_in_front_allowed;

static void gputchar(int c, int r, int g, int b)
{
  if (c == ' ' && (column == 0 || column == 80))
    if (!space_in_front_allowed) return;
  if (column == 80) {
    column = 0;
    line++;
  }
  while (line >= maxlines) {
    maxlines += 20;
    screen = realloc(screen, maxlines * 80 * sizeof(character));
    if (screen == NULL) abort();
    memset(screen + (maxlines-20) * 80, 0, 20 * 80 * sizeof(character));
  }
  if (c == '\n') {
    line++;
    column = 0;
    return;
  }
  screen[line*80+column].c = c;
  screen[line*80+column].r = r;
  screen[line*80+column].g = g;
  screen[line*80+column].b = b;
  column++;
}

static int wordlen(char *s)
{
  int l = 0;
  int v;
  while (*s && !isspace(*s)) {
    s += get_utf8(s, &v);
    l++;
  }
  return l;
}

static void cgprintf(int r, int g, int b, char *s, ...)
#ifdef __GNUC__
   __attribute__ ((format (printf, 4, 5)))
#endif
;

void cgprintf(int r, int g, int b, char *s, ...)
{
  int i;
  static char *buf = NULL;
  static int len = 0;
  int last_is_space = 0;
  int tlen;
  va_list ap;
  while (1) {
    va_start(ap, s);
    tlen = vsnprintf(buf, len, s, ap);
    va_end(ap);
    if (tlen < len) break;
    len += 1024;
    buf = realloc(buf, len * sizeof(character));
    if (buf == NULL) abort();
  }
  for (i = 0; buf[i];) {
    int v;
    if (!isspace(buf[i]) && last_is_space) {
      int l = wordlen(buf+i);
      if (l < 60 && column != 80 && l + column > 80)
        gputchar('\n', r, g, b);
    }
    i += get_utf8(buf+i, &v);
    gputchar(v, r, g, b);
    last_is_space = isspace(v);
  }
}

static void put_screen(void)
{
  character l[81];
  character *s;
  int i, j;

  ofull();
  for (i = 0; i < 22 && i+start_line < line; i++) {
    s = screen + (start_line + i) * 80;
    for (j = 0; j < 80 && s[j].c; j++) l[j] = s[j];
    l[j].c = 0;
    oputline(i, l);
  }
}

static void scroll_help(int delta)
{
  start_line += delta;
  if (start_line < 0) start_line = 0;
  if (start_line >= line) start_line = line-1;
  put_screen();
}

static void show_help(void)
{
  space_in_front_allowed = 1;
  column = 0;
  line = 0;
  start_line = 0;

  memset(screen, 0, maxlines * 80 * sizeof(character));
  cgprintf(128, 128, 255, "ESC/h: quit help     "
           "Up/Down/PageUp/PageDown: scroll text\n\n");

  cgprintf(180,180,255, "Commands:\n");
  //cgprintf(128,255,128, "    a");
  //cgprintf(180,180,255, "   aspect ratio\n");
  cgprintf(128,255,128, "    e");
  cgprintf(180,180,255, "   equalizer on/off\n");
  cgprintf(128,255,128, "    E");
  cgprintf(180,180,255, "   equalizer's controls\n");
  cgprintf(128,255,128, "    f");
  cgprintf(180,180,255, "   fullscreen on/off\n");
  cgprintf(128,255,128, "    F");
  cgprintf(180,180,255, "   set window size to current video size\n");
  //cgprintf(128,255,128, "    g");
  //cgprintf(180,180,255, "   channel's guide\n");
  //cgprintf(128,255,128, "    G");
  //cgprintf(180,180,255, "   global guide\n");
  cgprintf(128,255,128, "    h");
  cgprintf(180,180,255, "   help\n");
  //cgprintf(128,255,128, "    m");
  //cgprintf(180,180,255, "   mute/unmute\n");
  //cgprintf(128,255,128, "    M");
  //cgprintf(180,180,255, "   reset volume\n");
  cgprintf(128,255,128, "    q");
  cgprintf(180,180,255, "   quit\n");
  cgprintf(128,255,128, "    a");
  cgprintf(180,180,255, "   audio/video sync: audio back in time\n");
  cgprintf(128,255,128, "    A");
  cgprintf(180,180,255, "   audio/video sync: audio advance in time\n");
  //cgprintf(128,255,128, "    v");
  //cgprintf(180,180,255, "   volume down\n");
  //cgprintf(128,255,128, "    V");
  //cgprintf(180,180,255, "   volume up\n");
  cgprintf(128,255,128, "    DOWN arrow");
  cgprintf(180,180,255, "      volume down\n");
  cgprintf(128,255,128, "    UP arrow");
  cgprintf(180,180,255, "        volume up\n");
  //cgprintf(128,255,128, "    LEFT arrow");
  //cgprintf(180,180,255, "      previous channel\n");
  //cgprintf(128,255,128, "    RIGHT arrow");
  //cgprintf(180,180,255, "     next channel\n");
  //cgprintf(128,255,128, "    mouse's wheel");
  //cgprintf(180,180,255, "   volume up/down\n");
  //cgprintf(128,255,128, "    1/2/3/4/5/6/7/8/9/0");
  //cgprintf(180,180,255, "   set channel 1...10\n");
  //cgprintf(128,255,128, "    CTRL+g");
  //cgprintf(180,180,255, "   reload guide\n");

  cgprintf(180,180,255, "\nIn the equalizer's controls:\n");
  cgprintf(128,255,128, "    ESC");
  cgprintf(180,180,255, "   close the controls' window\n");
  cgprintf(128,255,128, "    MOUSE");
  cgprintf(180,180,255, "   use it to set the equalizer to your taste\n");

  gprintf("\n");
  cgprintf(128, 128, 255, "[END]\n");

  space_in_front_allowed = 0;

  put_screen();
}

void help_enter(void)
{
  show_help();
}

int help_key_press(int k)
{
  switch (k) {
  case KEY_UP: scroll_help(-1); break;
  case KEY_DOWN: scroll_help(1); break;
  case KEY_PAGE_UP: scroll_help(-10); break;
  case KEY_PAGE_DOWN: scroll_help(10); break;
  case 'h':
  case KEY_ESC: return OVER_QUIT;
  default: return OVER_NOTHING;
  }
  return OVER_REDRAW;
}
