#include "images.h"

#include <stdlib.h>
#include <pthread.h>

typedef struct image_t {
  struct image_t *next;
  struct image_t *last;
  char *data;
} image_t;

static pthread_cond_t cond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
static image_t *list;
static volatile int count;

char *get_image(void)
{
  image_t *r;
  char *ret;
  if (pthread_mutex_lock(&mutex)) abort();
  if (list == NULL) {
    if (pthread_mutex_unlock(&mutex)) abort();
    return NULL;
  }
  r = list;
  list = list->next;
  count--;
  if (list != NULL) list->last = r->last;
  ret = r->data;
  if (pthread_cond_signal(&cond)) abort();
  if (pthread_mutex_unlock(&mutex)) abort();
  free(r);
  return ret;
}

void put_image(char *image)
{
  image_t *r;
  r = calloc(1, sizeof(*r)); if (r == NULL) abort();
  r->data = image;
  r->last = r;
  if (pthread_mutex_lock(&mutex)) abort();
  while (count == 10)
    if (pthread_cond_wait(&cond, &mutex)) abort();
  if (list == NULL) list = r;
  else {
    list->last->next = r;
    list->last = r;
  }
  count++;
  if (pthread_cond_signal(&cond)) abort();
  if (pthread_mutex_unlock(&mutex)) abort();
}

int image_size(void)
{
  return count;
}

void reset_images(void)
{
  image_t *r;
  if (pthread_mutex_lock(&mutex)) abort();
  while (list != NULL) {
    r = list;
    list = list->next;
    free(r->data);
    free(r);
  }

  count = 0;

  if (pthread_cond_signal(&cond)) abort();
  if (pthread_mutex_unlock(&mutex)) abort();
}
