#include "msg.h"

#include <stdlib.h>
#include <string.h>
#include <pthread.h>

typedef struct msg_t {
  struct msg_t *last;
  struct msg_t *next;
  int id;
  arg_t a;
} msg_t;

typedef struct {
  pthread_cond_t cond;
  pthread_mutex_t mutex;
  msg_t *list;
} msg_queue_t;

static msg_queue_t queues[TO_LAST + 1];

void init_msg(void)
{
  int i;
  for (i = 0; i < TO_LAST + 1; i++) {
    pthread_cond_init(&queues[i].cond, NULL);
    pthread_mutex_init(&queues[i].mutex, NULL);
  }
}

void send_msg(int to, int msg_id, arg_t a)
{
  msg_t *m;
  msg_queue_t *q = &queues[to];

  m = calloc(1, sizeof(*m)); if (m == NULL) abort();
  m->id = msg_id;
  if (a != NULL)
    memcpy(m->a, a, sizeof(arg_t));
  m->last = m;

  if (pthread_mutex_lock(&q->mutex)) abort();

  if (q->list == NULL) q->list = m;
  else {
    q->list->last->next = m;
    q->list->last = m;
  }

  if (pthread_cond_broadcast(&q->cond)) abort();
  if (pthread_mutex_unlock(&q->mutex)) abort();
}

int get_msg(int to, arg_t a)
{
  int ret;
  msg_queue_t *q = &queues[to];
  msg_t *m;

  if (pthread_mutex_lock(&q->mutex)) abort();

  while (q->list == NULL)
    if (pthread_cond_wait(&q->cond, &q->mutex)) abort();

  m = q->list;
  if (q->list->next != NULL) q->list->next->last = q->list->last;
  q->list = q->list->next;

  ret = m->id;
  if (a != NULL)
    memcpy(a, m->a, sizeof(arg_t));
  free(m);

  if (pthread_mutex_unlock(&q->mutex)) abort();

  return ret;
}

int has_msg(int to)
{
  int ret;
  msg_queue_t *q = &queues[to];

  if (pthread_mutex_lock(&q->mutex)) abort();
  ret = q->list != NULL;
  if (pthread_mutex_unlock(&q->mutex)) abort();

  return ret;
}
