#include "sounds.h"

#include <stdlib.h>
#include <pthread.h>

typedef struct sound_t {
  struct sound_t *next;
  struct sound_t *last;
  char *data;
  int size;
} sound_t;

static pthread_cond_t cond = PTHREAD_COND_INITIALIZER;
static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
static sound_t *volatile list;
static volatile int count;

char *get_sound(int *ret_size)
{
  sound_t *r;
  int size;
  char *ret;
  if (pthread_mutex_lock(&mutex)) abort();
  if (list == NULL) {
    if (pthread_mutex_unlock(&mutex)) abort();
    return NULL;
  }
  r = list;
  list = list->next;
  count--;
  if (list != NULL) list->last = r->last;
  ret = r->data;
  size = r->size;
  if (pthread_cond_signal(&cond)) abort();
  if (pthread_mutex_unlock(&mutex)) abort();
  free(r);
  *ret_size = size;
  return ret;
}

void put_sound(char *sound, int size)
{
  sound_t *r;
  r = calloc(1, sizeof(*r)); if (r == NULL) abort();
  r->data = sound;
  r->size = size;
  r->last = r;
  if (pthread_mutex_lock(&mutex)) abort();
  while (count == 20)
    if (pthread_cond_wait(&cond, &mutex)) abort();
  if (list == NULL) list = r;
  else {
    list->last->next = r;
    list->last = r;
  }
  count++;
  if (pthread_cond_signal(&cond)) abort();
  if (pthread_mutex_unlock(&mutex)) abort();
}

void reset_sounds(void)
{
  sound_t *r;
  if (pthread_mutex_lock(&mutex)) abort();
  while (list != NULL) {
    r = list;
    list = list->next;
    free(r->data);
    free(r);
  }

  count = 0;

  if (pthread_cond_signal(&cond)) abort();
  if (pthread_mutex_unlock(&mutex)) abort();
}
