/* TODO: API cleanup, this was taken from another project of mine with minor
 * changes
 * I don't like PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP at all
 */

#include "text_over.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdint.h>
#define __USE_GNU
#include <pthread.h>

#include "font.h"
#include "util.h"
#include "msg.h"

static unsigned char screen[640*352*4];
static character tscreen[22][80];
static int line, column;
static int tscreen_dirty;
static int tscreen_is_clear;

static pthread_mutex_t lock = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;

#define LOCK do { if (pthread_mutex_lock(&lock)) abort(); } while (0)
#define UNLOCK do { if (pthread_mutex_unlock(&lock)) abort(); } while (0)

static void set_text(void)
{
  int line = 0;
  int col  = 0;
  int pixel = 0;
  int i, j;
  memset(screen, 0, 640*352*4);
  tscreen_is_clear = 1;
  for (line = 0; line < 22; line++)
  for (col = 0; col < 80; col++) {
    if (tscreen[line][col].c == 0) continue;
    tscreen_is_clear = 0;
    font_elem *e = get_font_elem(tscreen[line][col].c);
    for (i = 0; i < 16; i++) {
      int v = e->c[i];
      pixel = i * 640 + col * 8 + line * 16 * 640;
      for (j = 0; j < 8; j++, v <<= 1, pixel++) {
        if (v & 0x80) {
          screen[pixel * 4 + 0] = tscreen[line][col].r;
          screen[pixel * 4 + 1] = tscreen[line][col].g;
          screen[pixel * 4 + 2] = tscreen[line][col].b;
          screen[pixel * 4 + 3] = 255;
        } else {
          screen[pixel * 4 + 0] =
          screen[pixel * 4 + 1] =
          screen[pixel * 4 + 2] = 0;
          screen[pixel * 4 + 3] = 230; /* transparency, change not good */
        }
      }
    }
  }
}

char *get_text_over(void)
{
  LOCK;
  if (tscreen_dirty) { set_text(); tscreen_dirty = 0; }
  if (tscreen_is_clear) { UNLOCK; return NULL; }
  UNLOCK;
  return (char *)screen;
}

void oputchar(int c, int r, int g, int b)
{
  LOCK;
  tscreen_dirty = 1;
  if (column == 80) {
    column = 0;
    line++;
  }
  if (line == 22) {
    memmove(tscreen, &tscreen[1][0], 21*80*sizeof(character));
    memset(&tscreen[21][0], 0, 80*sizeof(character));
    line = 21;
    column = 0;
  }
  if (c == '\n') {
    line++;
    column = 0;
    UNLOCK;
    return;
  }
  tscreen[line][column].c = c;
  tscreen[line][column].r = r;
  tscreen[line][column].g = g;
  tscreen[line][column].b = b;
  column++;
  UNLOCK;
}

void oup(void)
{
  LOCK;
  line--;
  if (line < 0) line = 0;
  memset(&tscreen[line][0], 0, 80*sizeof(character));
  tscreen_dirty = 1;
  UNLOCK;
}

void oprintf(char *s, ...)
{
  int i;
  static char *b = NULL;
  static int len = 0;
  int tlen;
  va_list ap;
  LOCK;
  while (1) {
    va_start(ap, s);
    tlen = vsnprintf(b, len, s, ap);
    va_end(ap);
    if (tlen < len) break;
    len += 1024;
    b = realloc(b, len);
    if (b == NULL) abort();
  }
  for (i = 0; b[i];) {
    int v;
    i += get_utf8(b+i, &v);
    oputchar(v, 255, 255, 255);
  }
  UNLOCK;
}

void oputline(int l, character *s)
{
  int i;
  LOCK;
  for (i = 0; s[i].c; i++) {
    tscreen[l][i].c = s[i].c;
    tscreen[l][i].r = s[i].r;
    tscreen[l][i].g = s[i].g;
    tscreen[l][i].b = s[i].b;
  }
  UNLOCK;
}

void coprintf(int r, int g, int b, char *s, ...)
{
  int i;
  static char *buf = NULL;
  static int len = 0;
  int tlen;
  va_list ap;
  LOCK;
  while (1) {
    va_start(ap, s);
    tlen = vsnprintf(buf, len, s, ap);
    va_end(ap);
    if (tlen < len) break;
    len += 1024;
    buf = realloc(buf, len);
    if (buf == NULL) abort();
  }
  for (i = 0; buf[i];) {
    int v;
    i += get_utf8(buf+i, &v);
    oputchar(v, r, g, b);
  }
  UNLOCK;
}

void coputline(int r, int g, int b, int l, int *s)
{
  int i;
  LOCK;
  for (i = 0; s[i]; i++) {
    tscreen[l][i].c = s[i];
    tscreen[l][i].r = r;
    tscreen[l][i].g = g;
    tscreen[l][i].b = b;
  }
  UNLOCK;
}

void ofull(void)
{
  int i, j;
  LOCK;
  tscreen_dirty = 1;
  line = column = 0;
  for (i = 0; i < 22; i++) for (j = 0; j < 80; j++) tscreen[i][j].c = ' ';
  UNLOCK;
}

void oclear(void)
{
  LOCK;
  tscreen_dirty = 1;
  line = column = 0;
  memset(tscreen, 0, 80*22*sizeof(character));
  UNLOCK;
}

void odelay(int x)
{
  arg_t a;
  a[0].i = new_id();
  send_msg(TO_APP, M_TEXT_OVER_ON, a);
  new_timer(x, TO_APP, M_TEXT_OVER_OFF, a);
}
