#define _GNU_SOURCE
#include "thread.h"

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <sched.h>

#include "util.h"

typedef struct {
  void *(*f)(void *);
  void *b;
  int rt;
} wrapper_t;

static void use_rt_scheduler(void)
{
  struct sched_param p;
  p.sched_priority = sched_get_priority_min(SCHED_FIFO);
  if (sched_setscheduler(0, SCHED_FIFO, &p)) {
    fprintf(stderr, "error: could not set realtime for video, expect jitter\n");
    fprintf(stderr, "run the program as root (not a good idea)\n");
    fprintf(stderr, "or as root do: setcap cap_sys_nice+ep vlb)\n");
    exit(1);
  }
}

static void *wrapper(void *_w)
{
  wrapper_t *w = _w;
  void *(*f)(void *) = w->f;
  void *b = w->b;
  int rt = w->rt;
  free(w);

  if (rt) use_rt_scheduler();
  remove_capabilities();

  return f(b);
}

static pthread_t _new_thread(void *(*f)(void *), void *b, char *name)
{
  pthread_t t;
  pthread_attr_t att;

  if (pthread_attr_init(&att))
    { fprintf(stderr, "pthread_attr_init err\n"); exit(1); }
#if 0
  if (pthread_attr_setdetachstate(&att, PTHREAD_CREATE_DETACHED))
    { fprintf(stderr, "pthread_attr_setdetachstate err\n"); exit(1); }
#endif
  if (pthread_create(&t, &att, f, b))
    { fprintf(stderr, "pthread_create err\n"); exit(1); }
  if (pthread_attr_destroy(&att))
    { fprintf(stderr, "pthread_attr_destroy err\n"); exit(1); }

  if (pthread_setname_np(t, name))
    { fprintf(stderr, "pthread_setname_np err name '%s'\n", name); exit(1); }

  return t;
}

pthread_t new_thread(void *(*f)(void *), void *b, char *name)
{
  wrapper_t *w = malloc(sizeof(*w)); if (w == NULL) abort();
  w->f = f;
  w->b = b;
  w->rt = 0;
  return _new_thread(wrapper, w, name);
}

pthread_t new_rt_thread(void *(*f)(void *), void *b, char *name)
{
  wrapper_t *w = malloc(sizeof(*w)); if (w == NULL) abort();
  w->f = f;
  w->b = b;
  w->rt = 1;
  return _new_thread(wrapper, w, name);
}

void detach_thread(void)
{
  pthread_detach(pthread_self());
}

void thread_wait(pthread_t t)
{
  if (pthread_join(t, NULL)) abort();
}
