#include "util.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/wait.h>
#include <unistd.h>
#include <errno.h>
#include <linux/capability.h>
#include <sys/syscall.h>

#include "msg.h"
#include "thread.h"

uint64_t get_time(void)
{
  struct timespec t;
  if (clock_gettime(CLOCK_MONOTONIC_RAW, &t)) abort();
  return (uint64_t)t.tv_sec * 1000 + (uint64_t)t.tv_nsec / 1000000;
}

void *process_watch(void *_)
{
  while (1) {
    int status;
    int process;
    process = wait(&status);
    if (process == -1) {
      if (errno != ECHILD) { perror("wait"); exit(1); }
      sleep(2);
      continue;
    }
    arg_t a;
    a[0].i = process;
    a[1].i = status;
    send_msg(TO_APP, M_PROCESS_DIED, a);
  }
}

/****************************************************************************/
/* timer code begin                                                         */
/* (we could use only one thread and a list, but I'm lazy)                  */
/****************************************************************************/

typedef struct {
  int timeout_ms;
  int to;
  int msg_id;
  arg_t a;
} my_timer_t;

void *timer_thread(void *_t)
{
  my_timer_t *t = _t;
  detach_thread();
  usleep(t->timeout_ms * 1000);
  send_msg(t->to, t->msg_id, t->a);
  free(t);
  return 0;
}

void new_timer(int timeout_ms, int to, int msg_id, arg_t a)
{
  my_timer_t *t = malloc(sizeof(*t)); if (t == NULL) abort();
  t->timeout_ms = timeout_ms;
  t->to = to;
  t->msg_id = msg_id;
  if (a != NULL)
    memcpy(t->a, a, sizeof(arg_t));
  else
    memset(t->a, 0, sizeof(arg_t));
  new_thread(timer_thread, t, "timer");
}

/****************************************************************************/
/* timer code end                                                           */
/****************************************************************************/

int get_utf8(char *s, int *v)
{
#define V do { *v = *n; return 1; } while (0)
  int len;
  int val;
  unsigned char *n = (unsigned char *)s;

  if (*n == 0xc0 || *n == 0xc1 || *n >= 0xf5) V;
  if (!(*n & 0x80)) V;
  else
  if ((*n & 0xe0) == 0xc0) { /* 110xxxxx */
    if ((*(n+1) & 0xc0) != 0x80) V;
    val = ((*n & 0x1f) << 6) | (*(n+1) & 0x3f);
    if (val < 0x80 || val > 0x7ff) V;
    *v = val;
    len = 2;
  } else
  if ((*n & 0xf0) == 0xe0) { /* 1110xxxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80) V;
    val = ((*n & 0x0f) << (6*2)) | ((*(n+1) & 0x3f) << 6) | (*(n+2) & 0x3f);
    if (val < 0x800 || val > 0xffff) V;
    *v = val;
    len = 3;
  } else
  if ((*n & 0xf8) == 0xf0) { /* 11110xxx */
    if ((*(n+1) & 0xc0) != 0x80 ||
        (*(n+2) & 0xc0) != 0x80 ||
        (*(n+3) & 0xc0) != 0x80) V;
    val = ((*n & 0x07) << (6*3)) | ((*(n+1) & 0x3f) << (6*2)) |
          ((*(n+2) & 0x3f) << 6) | (*(n+3) & 0x3f);
    if (val < 0x10000 || val > 0x10ffff) V;
    *v = val;
    len = 4;
  } else V;
  /* we don't deal with values > 0xffff */
  if (val > 0xffff) V;
  return len;
#undef V
}

/* return a short-term ID */
int new_id(void)
{
  static int id = 0;
  int ret = __atomic_fetch_add(&id, 1, __ATOMIC_SEQ_CST);
  __atomic_and_fetch(&id, 65535, __ATOMIC_SEQ_CST);
  return ret & 65535;
}

void remove_capabilities(void)
{
  struct __user_cap_header_struct hdr;
  struct __user_cap_data_struct cap[2];
  hdr.version = _LINUX_CAPABILITY_VERSION_3;
  hdr.pid = 0;
  if (syscall(SYS_capget, &hdr, cap) == -1) { perror("capget"); exit(1); }
  cap[0].effective &= ~(1 << CAP_SYS_NICE);
  cap[0].permitted &= ~(1 << CAP_SYS_NICE);
  cap[0].inheritable &= ~(1 << CAP_SYS_NICE);
  if (syscall(SYS_capset, &hdr, cap) == -1) { perror("capset"); exit(1); }
}
