#include "video.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sched.h>

#include "images.h"
#include "msg.h"
#include "util.h"

typedef struct {
  uint64_t start;
  uint64_t frame;
  float fps;
  int rt_mode;
} video_player_internal_t;

video_player_t *init_video_player(uint64_t start, float fps, int rt_mode)
{
  video_player_internal_t *ret;
  ret = calloc(1, sizeof(*ret)); if (ret == NULL) abort();
  ret->start = start;
  ret->fps = fps;
  ret->rt_mode = rt_mode;
  return ret;
}

void *video_player(video_player_t *_v)
{
  video_player_internal_t *v = _v;
  char *image;
  uint64_t cur_time;
  uint64_t next_time;
  uint64_t delay;
  arg_t a;
  int paused = 1;
  int end_id;
  int end_reached = 0;

  while (1) {
    if (paused || has_msg(TO_VIDEO)) {
      arg_t a;
      int msg = get_msg(TO_VIDEO, a);
      switch (msg) {
      case M_PAUSED: paused = 1; break;
      case M_UNPAUSED: paused = 0; v->start = a[0].u; break;
      case M_SEEK: paused = 0; v->frame = 0; break;
      case M_WAITING_END: end_id = a[0].i; end_reached = 1; break;
      }
    }

    if (paused) continue;

    image = get_image();
    if (image == NULL) {
      printf("no image!\n");
      if (end_reached) {
        arg_t a;
        /* wait 2 frames */
        usleep(2 * 1000000 / v->fps);
        a[0].i = end_id;
        send_msg(TO_APP, M_END_REACHED, a);
        end_reached = 0;
        /* unpause audio if needed */
        if (v->frame == 0) {
          a[0].u = v->start;
          send_msg(TO_AUDIO, M_UNPAUSED, a);
        }
      }
      usleep(1000);
      continue;
    }
    a[0].p = image;
    send_msg(TO_X, M_SHOW_IMAGE, a);
    if (get_msg(TO_VIDEO_FROM_X, 0) != M_CONSUMED) abort();
    /* set start time only after getting the first image, better realtime behavior */
    if (v->frame == 0) {
      arg_t a;
      v->start = get_time();
      /* also unpause audio (useful after a SEEK, avoids bad audio) */
      a[0].u = v->start;
      send_msg(TO_AUDIO, M_UNPAUSED, a);
    }
    v->frame++;
    cur_time = get_time();
    next_time = v->start + (uint64_t)(v->frame * 1000 / v->fps);  /* TODO: debug why (uint64_t) cast is needed */
    delay = next_time - cur_time;
    if (cur_time > next_time) { printf("delay 0 images size %d\n", image_size()); delay = 0; }
    /* don't wait too much or too low */
    if (delay > 1000*2/v->fps) { printf("too high delay\n"); delay = 1000*2/v->fps; }
    if (delay < 1000/2/v->fps) delay = 1000/2/v->fps;
    if (v->rt_mode == RT_BUSY_WAIT) {
      uint64_t dd;
      do {
        dd = get_time() - cur_time;
      } while (dd < delay);
    } else {
      usleep(delay * 1000);
      uint64_t dd = get_time() - cur_time;
      if (dd > delay + 10) printf("oversleep %ld delay %ld\n", dd, delay);
    }
  }

  return 0;
}

uint64_t video_current_frame(video_player_t *_v)
{
  video_player_internal_t *v = _v;
  return v->frame;
}
